/* XPM */
static const char *_logo_schism_xpm[] = {
/* width height ncolors const chars_per_pixel */
"290 50 3 1",
/* colors */
"  c #000000",
". c #FFFFFF",
"X c None",
/* pixels */
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXX.........................................................XXXX....XXXX....................................................................................................................................XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXX..........................................................XXXX...XXXXX....................................................................................................................................XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXX..XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXX.XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX....................XXXXXX.............XXXXXXXX.XXXXXXXX....XXXXXX..XXXXXX.................XXXXXXX..................XXXXXXXXXXXX....XXXXXXX............XXXXXXX..............XXXXXX................XXXXXX....XXXXXXXXXXX......XXXXXXXXXXXXX..................XXXXXXX....................XX",
"XXXXXXXXX....................XXXXX.............XXXXXXXX..XXXXXXXX....XXXXX...XXXXX.................XXXXXXX...................XXXXXXXXXXXX....XXXXXX............XXXXXXX...............XXXXX................XXXXXXX....XXXXXXXXXX......XXXXXXXXXXXXX...................XXXXXX....................XXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX.....XXXXXXXXXXXXXXXX...XXXXXXXX....XXXX....XXXX.....XXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXXXXXXXXXX....XXXXX....XXXXXXXXXXXXXXX.....XXXXXXX....XXXX.....XXXXXXXXXXXXXXXXXXX....XXXXXXXXX......XXXXXXXXXXXXX.....XXXXXXXXXXX....XXXXX....XXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXX......XXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXX......XXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXX......XXXXXXXXXXXXXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX................XXXX....XXXX....................XXXX....XXXX....XXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX................XXXX....XXXXXXXXXXXXXXXXXXXX............................XXXX....XXX............XXXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX................XXXX....XXXX....................XXXX....XXXX....XXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX................XXXX....XXXXXXXXXXXXXXXXXXXX............................XXXX....XX............XXXXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXX....XXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXX....XXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXX....XXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXX...XXXXX....XXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXX....XXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXX....XXXX....XXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXX",
"............................XXXXX...............XXXXX....XXXXXXXX....XXXX....XXXXX..................XXXXX...XXXXX..XXXXXX....XXXXXXXXXXXX...XXXXX...XXXXXXXXXXXXXXXXX...XXXXXXXXX....XXXX..................XXXXXX...XXXXXXXXXXXXXXXXXXXXX....XXXX..................XXXXXX...XXXXXXXXXXXXXXXXXXXXXX",
"...........................XXXXXX..............XXXXXX....XXXXXXXX....XXXX....XXXX..................XXXXXX..XXXXXXXXXXXXXX....XXXXXXXXXXXX..XXXXXX..XXXXXXXXXXXXXXXXXX..XXXXXXXXXX....XXXX.................XXXXXXX..XXXXXXXXXXXXXXXXXXXXXX....XXXX.................XXXXXXX..XXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXX...XXXXXXXX....XXXX.X..XXXXXXXXXXXX..X..XX..XXXX....XXX....XXXXXXXX....XXXX.X..XXXXXXXXXX...XX...X..XXXX....XXX....XX.X.XXXX...XXXX.X..XX.X.X..XXX..XX...X..XXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX...XX..XXX....XXX...XXXXXXXXX.XX.XXX.X..XXXXXXXXXXXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXXX.XXX.X.XX.XXXX....XXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXX.XX.X.X.XX.XXXX....XXX.XXXX.....XX...XXXX.X.X.XX.X.X.XXX.XXXX.X.XXX.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXX.XX.X.X.X.XXXXX....XXX.XX.XXXXXXXX.XX.XX.X.X.XXXXXXXXXXXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXX.XXXXX....XXXXX.XXXX.XX.XXX.X.X.XXXX..X.X.XX.XXXX....XXX...XXX....XXXXX.XXXX.X.XXXX.X.X.XX.XX.X.X.XX.XXXX....XXX...XX.....XX.XX.XXX.X.X.XX...XX.XX.X.XX.X.XX.XXXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XX....X.X.X...XXX....XXX.XX.XX....XX....XX.X.X.XXXX.X.X.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXX....XXXXXXXX...XXXXX.XX.XXXXXXXXXXX...X..XX...XXX....XXX....XXXXXXXX...XXXXX.X..XXXXXXXXXX...XX...X...XXX....XXX.XXXXX.X.XXX....XXXX.XX.XXXX.X..XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX.XX.X...XX..XXX....XXX...XXXXXXXXXXXX.XXX.X...XXXXXXXXXXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX....XXXX.X..XXXXXXXXXX...XX...XX..XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X..XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX..XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXXX..XX.X.X.X.XXX....XXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXX.XX.X.X.X..XXXX....XXXX..XXXX..XXXX..XXXXXXXXXXX.XXX..XX.XXXX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXX.XX.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXX...XXX....XXXXX.XXXX.X.XXXX.X.X.XX.XX.X.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX..XX.XX.X.XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.XXX.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX...XXXXX.X..XXXXXXXXXX...XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...XX.XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...XX.XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX....XXXX.X..XXXXXXXXXX...XX...X.X.XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX..X..XXX..XX...X...XXX....XXX....XX.X.XXXX...XXXX.X..XX..XX..XXX..XX...X..XXXX....XXX...XXXXXXXXX.XX.XXX.X..XXXXXXXXXXXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXXX..XX.X.X.X.XXX....XXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXX.XX.X.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXX.XXX..XX.XXXX.X.X.X.XXX....XXX.XXXX.....XX...XXXX.X.X.XXXX.X..XX.XXXX.X.XXX.XXX....XXX.XX.XXXXXXXX.XX.XX.X.X.XXXXXXXXXXXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXX...XXX....XXXXX.XXXX.X.XXXX.X.X.XX.XX.X.X.X...XXX....XXXX..XXXX..XXXX..XXXX.X.X.XX...X..XX.X.XX.X.X.X.XXX....XXX...XX.....XX.XX.XXX.X.X.XXX.XX..XX.X.XX.X.XX.XXXX....XXX.XX.XX....XX....XX.X.X.XXXX.X.X.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX...XXXXX.X..XXXXXXXXXX...XX...XXX.XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX..X..XXX..XX...X...XXX....XXX.XXXXX.X.XXX....XXXX.XX.XX...XX.XXX..XX...X...XXX....XXX...XXXXXXXXXXXX.XXX.X...XXXXXXXXXXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX....XXXX.X..XXXXXXXXXX...XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX..XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX..XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXXX..XX.X.X.X.XXX....XXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXX.XX.X.X.X.XXXXX....XXXX..XXXX..XXXX..XXXXXXXXXXX..XX..XXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXX.X.XXX.XXXX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXX.XX.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXX...XXX....XXXXX.XXXX.X.XXXX.X.X.XX.XX.X.X.XX..XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXXX.X..XXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.XXX.XX.X.XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.XXX.XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX...XXXXX.X..XXXXXXXXXX...XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX.XXX.XXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X.XXXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...XX.XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...XX.XXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX....XXXX.X..XXXXXXXXXX...XX...X..XXXX....XXX....XXXXXXXXX...XXXX.X..XXXXXXXXXXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX..XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX..XX..XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXXXXXXXXXXXXXXXXXX..XX.X.X.X.XXX....XXX.XXXXXXXXXXX...XXXXX.XX.XXXXXXXXXX.XX.X.X.XX.XXXX....XXX.XXXXXXXXXXX...XXXX.X.X.XXXXXXXXXXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXXX.X.XXX.XXXX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXXXXXXXXXX.XX..XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXX..XXXX..XXXX..XXXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXX...XXX....XXXXX.XXXX.X.XXXX.X.X.XX.XX.X.X.XX.XXXX....XXX...XXX....XX.XX.XXX.X.X.XXX.X.X.XXX..XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.XX..XX.X.XX.X.X.X.XXX....XXXX..XXXX..XXXX..XXXX.X.X.XXX.XX..XXX..XX.X.X.X.XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X...XXX....XXX....XXXXXXXX...XXXXX.X..XXXXXXXXXX...XX...X...XXX....XXX....XXXXXXXX....XXXX.XX.XXXXXXXXXXXXXXX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X..XXX..XX...X...XXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXX...X..XXXXXXX...X...XXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
};
