/*
 * Schism Tracker - a cross-platform Impulse Tracker clone
 * copyright (c) 2003-2005 Storlek <storlek@rigelseven.com>
 * copyright (c) 2005-2008 Mrs. Brisby <mrs.brisby@nimh.org>
 * copyright (c) 2009 Storlek & Mrs. Brisby
 * copyright (c) 2010-2012 Storlek
 * URL: http://schismtracker.org/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef SCHISM_VERSION_H_
#define SCHISM_VERSION_H_

#include "util.h" // PURE

/* various boilerplate defined in version.c */
extern const char *ver_short_copyright;
extern const char *ver_short_based_on;
extern uint16_t ver_cwtv; /* lower 12 bits of the IT/S3M cwtv field */
extern uint32_t ver_reserved; /* full version number in case 12 bits are not enough */

SCHISM_PURE extern const char *schism_banner(int classic);

/* need to call this at startup */
void ver_init(void);

/* get yyyy-mm-dd or 0.nn version from cwtv + reserved (buf should be >=11 chars) */
void ver_decode_cwtv(uint16_t cwtv, uint32_t reserved, char buf[11]);

/* get a version number from a given date */
SCHISM_CONST uint32_t ver_mktime(uint32_t year, uint32_t month, uint32_t day);

int ver_to_date(uint32_t ver, uint32_t *py, uint32_t *pm, uint32_t *pd);

/* "20090603" */
int ver_parse_schism_version(const char *ver, uint32_t *pyear, uint32_t *pmonth, uint32_t *pday);
/* "Wed Jun  3 23:39:19 2009" */
int ver_parse_ctimestamp(const char *timestamp, uint32_t *pyear, uint32_t *pmonth, uint32_t *pday);
/* "Jun  3 2009" */
int ver_parse_cdate(const char *cdate, uint32_t *pyear, uint32_t *pmonth, uint32_t *pday);

#endif /* SCHISM_VERSION_H_ */

