/*
 * Schism Tracker - a cross-platform Impulse Tracker clone
 * copyright (c) 2003-2005 Storlek <storlek@rigelseven.com>
 * copyright (c) 2005-2008 Mrs. Brisby <mrs.brisby@nimh.org>
 * copyright (c) 2009 Storlek & Mrs. Brisby
 * copyright (c) 2010-2012 Storlek
 * URL: http://schismtracker.org/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "headers.h"

#include "mem.h"

void *mem_alloc(size_t amount)
{
	void *q = malloc(amount);
	SCHISM_RUNTIME_ASSERT(q, "Failed to allocate heap memory.");
	return q;
}

void *mem_calloc(size_t nmemb, size_t size)
{
	void *q = calloc(nmemb, size);
	SCHISM_RUNTIME_ASSERT(q, "Failed to allocate initialized heap memory.");
	return q;
}

void *mem_realloc(void *orig, size_t amount)
{
	void *q;

	if (!orig)
		return mem_alloc(amount);

	q = realloc(orig, amount);
	SCHISM_RUNTIME_ASSERT(q, "Failed to reallocate heap memory.");
	return q;
}

char *strn_dup(const char *s, size_t n)
{
	char *q = mem_alloc((n + 1) * sizeof(char));
	memcpy(q, s, n * sizeof(char));
	q[n] = '\0';
	return q;
}

char *str_dup(const char *s)
{
	return strn_dup(s, strlen(s));
}
