/* XPM */
static const char *_logo_it_xpm[] = {
/* width height ncolors const chars_per_pixel */
"290 50 4 1",
/* colors */
"  c #000000",
". c #8CA8BC",
"X c #FFFFFF",
"o c None",
/* pixels */
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooooXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooXXXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooXXXXXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooXXXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXooooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooXXXXXXXXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXoooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooXXXXXXXXXXooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXXXXXXXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooXXXXXXXXXoooooXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXXXXXXXXXXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooXXXXXXXXXoooooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXXXXXXoooooooooooXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooXXXXXXXXoooooooXXXooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooXXXoXXXooooooooooooooooooooooooooooooooooooooooooooXXXXXXXXXoooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooXXXXXXXXoooooooXXXXoooooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooooooooXXXXoXXoooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooXXXXXXXooooooooXXXXooooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooooooooXXXXoXXXooooooooooooooooooooooooooooooooooooooooooooooXXXXXXXooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooXXXXXXXooooooooXXXXooooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooooooooXXXXoXXXooooooooXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooXXXXXXoooooooooXXXXoooooooooooooooooooooooooooooooooooXXXXXooooooooooooooooooooooooXXoooooXXXooXXXoooooooXXXXoooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXXXoooooooooooooooooXXXooooooooooooooooooooooooooooo",
"oooooooooooXXXXXXXXXoooooXXXXooooooooooooooXXXXooooooXXooooooooooXXXXoooooooooooooooXXXooooooXXXooooXXXXoXXXooooooooXXXXoooooooooooXXXXXXXoooooooooooooooooooooooooooooooooooooXXXXoooooXXooooooooooooooooooooXXXXXooooooooooooXXXXXXoooooXXXXXXoooooXXooooooooXXXXXXXXooooXXooooooooooooooooooooo",
"oooooooooooXXXXXXXXooooooXXXXooooooXXoooooXXXXXoooooXXXoooooooooXXXXooooXXXXoooooooXXXXXooooXXXXXoooXXXoXXXooooooooXXXXXoooooooooXXXXXXXXXooooooooooooooooooooooooooooooooooooXXXXXooooXXXXXXXXXXooooooooooXXXXXXXXXoooooooooXXXXXXXXooooXXXXXXooooXXXXXooooooXXXXXXXXXoooXXXXXXXXXooooooooooooooo",
"ooooooooooooXXXXXXooooooXXXXoooooXXXXXoooXXXXXXooooXXXXXoooooooXXXXXooXXXXXXooooooXXXXXooooXXXXXoooXXXXXXXooooooooXXXXXXooooooooXXXXXoXXXXoooooooooooooooooooooooooooooooooooXXXXXooooXXXXXXXXXXXXoooooooXXXXXXXXXXXoooooooXXXXXXXXXoooooXXXXXoooXXXXXXXoooooXXXXXoXXXXooXXXXXXXXXXXoooooooooooooo",
"ooooooooooooooXXooooooooXXXoooooXXXXXoooXXXXXXXoooXXXXXXooooooXXXXXoXXXXXXXXoooooXXXXXooooXXXXXXoooXXXXXXooooooooXXXXXXXXooooooXXXXoooXXXXoooooooooooooooooooooooooooooooooooXXXXooooXXXXXXXXXXXXXoooooXXXXXXXXXXXXXooooooXXXXXXXXXXooooXXXXXoooXXXXXXXXooooXXXXoooXXXooXXXXXXXXXXXXoooooooooooooo",
"oooooooooooooooooooooooXXXXooooXXXXXXoXXXXXXXXoooXXXXXXooooooXXXXXXXXXXXXXXXooooXXXXXooooXXXXXXoooXXXXXXooooooooXXXoXXXXXoooooXXXXoooXXXXoooooooooooooooooooooooooooooooooooXXXXXoooXXXXXXXXXXXXXoooooXXXXXXXXXXXXXooooooXXXXXXXXXXooooXXXXXXoXXXXXoXXXXoooXXXXoooXXXooXXXXXXXXXXXXooooooooooooooo",
"oooooooooooooooooooooooXXXooooXXXXXXoXXXXXXXXXoXXXXXXXooooooXXXXXXXXXoXXXXXooooXXXXXooooXXXXXXooooXXXXXooooooooXXXooXXXXXooooXXXXooXXXXXoooooooooooooooooooooooooooooooooooXXXXXooooXXXXXXXXXXXXoooooXXXXXXXoXXXXXXoooooXXXXXXoXXXoooooXXXXXoXXXXooXXXXoooXXXXooXXXXoooXXXXXXXXXXXoooooooooooooooo",
"ooooooooooooooooooooooXXXXooooXXXXXoXXXXXXXXXoXXXXXXXooooooXXXXXXXXXoooXXXXooooXXXXooooXXXXXXoooooXXXXooooooooXXXoooXXXXoooooXXXXXXXXXooooooooooooooooooooooooooooooooooooXXXXXXooooXXXooooXXXXoooooXXXXXooooXXXXXXooooXXXXXXooXXoooooXXXXXoXXXooooXXXooooXXXXoXXXXooooXXXoooXXXXooooooooooooooooo",
"oooooooooooooooooooooXXXXooooXXXXXoXXXXoXXXXoXXXXXXXXoooooXXoXXXXXXoooXXXXooooXXXXooooXXXXXXooooooXXXooooooooXXXooooXXXXooooXXXXXXXXoooooooXooooooooooooooooooooooooooooooXXXXXooooooXooooXXXXoooooXXXXXooooXXXXXXoooooXXXXooooooooooXXXXXXXXXooooXXXooooXXXXXXXXooooooXXoooXXXXoooooooooooooooooo",
"oooooooooooooooooooooXXXXooooXXXXooXXXoXXXXXXXXXoXXXoooooXXXoXXXXXoooXXXXooooXXXXooooXXXXXXXoooooXXXXoooooooXXXoooooXXXooooXXXXXXXooooooooXXoooooooooooooooooooooooooooooXXXXXXoooooXXooooXXXoooooXXXXXooooXXXXXXoooooXXXXXoooooooooXXXXXXXXXoooXXXooooooXXXXXXoooooooXXXoooXXXooooooooooooooooooo",
"ooooooooooooooooooooXXXXoooooXXXXoXXXooXXXXXXXXoXXXoooooXXXoXXXXXooooXXXooooXXXXXoooXXXXXXXooooXXXXXXooooooXXXXXoooXXXXoooXXXXXXXoooooooXXXooooooooooooooooooooooooooooooXXXXXoooooXXooooXXXXooooXXXXXooooXXXXXXoooooXXXXXoooooooooXXXXXXXXXXXXXXXoooooXXXXXXooooooooXXXoooXXXXooooooooooooooooooo",
"oooooooooooooooooooXXXXooooooXXXXXXXooXXXXXXXXooXXXoooXXXXooXXXXXXooXXXoooXXXXXXXoXXXXXXXXXoooXXXXXXXXooooXXXXXXXooXXXooXXXooXXXXooooooXXXooooooooooooooooooooooooooooooXXXXXXoooooXoooooXXXXoooXXXXXooooXXXoXXXooooXXXXXXoooooooXXXXXXXXXXXXXXXXoooooXXXXXXXoooooooXXXooooXXXXoooXXXooooooooooooo",
"oooooooooooooooooooXXXXooooooXXXXXXoooXXXXXXooXXXXXXXXXXXooXXXXXXXXXXXXXXXXXoXXXXXXXXXoXXXXXXXXXooXXXXXoXXXXXXXXXXXXoooXXXooXXXXXXooXXXXXoooooooooooooooooooooooooooooooXXXXXoooooooooooXXXXXoXXXXXXXoooXXXooXXXXXXXXXoXXXXooooXXXXXoXXXXXoXXXXXXooooXXXoXXXXXoooXXXXXooooXXXXXoXXXXXooooooooooooo",
"ooooooooooooooooooXXXXoooooooXXXXXXoooXXXXXXooXXXXXXXXXooooXXXXXXXXXXXXXXXXoooXXXXXXXoXXXXXXXXXoooXXXXXXXXXoXXXXXXXXXXXXXooooXXXXXXXXXXXoooooooooooooooooooooooooooooooXXXXXXoooooooooooXXXXXXXXXoXXXXXXXXooXXXXXXXXXooXXXXXXXXXXXXooXXXXoooXXXXXXoXXXXooXXXXXXXXXXXXoooooXXXXXXXXXooooooooooooooo",
"oooooooooooooooooXXXXXoooooooXXXXXoooXXXXXXoooXXXXXXXXooooXXXXXXXXXXXXXXXoooooXXXXXXooXXXXXXXXooooXXXXXXXXoooXXXXXXXXXXXoooooXXXXXXXXXooooooooooooooooooooooooooooooooXXXXXXooooooooooooXXXXXXXooooXXXXXXoooXXXXXXXXoooXXXXXXXXXXXooooXXXoooXXXXXXXXXXoooXXXXXXXXXXoooooooXXXXXXXooooooooooooooooo",
"oooooooooooooooooXXXXXooooooooXXXooooXXXXoooooXXXXXXoooooXXXXXXXXXXXXXXooooooooXXXoooooXXXXXXooooooXXXXXXoooooXXXXXXXXooooooooXXXXXXXoooooooooooooooooooooooooooooooooXXXXXXoooooooooooooXXXXXooooooXXXXooooXXXXXXooooooXXXXXXXXooooooXXoooooXXXXXXXooooooXXXXXXXooooooooooXXXXXoooooooooooooooooo",
"oooooooooooooooooXXXXXoooooooooXooooooXXoooooooXXXoooooooXXXXoooXXXXXooooooooooooooooooXXXooooooooooooooooooooooXXXXooooooooooooXXooooooooooooooooooooooooooooooooooooXXXXXoooooooooooooooXXoooooooooXXoooooXXXXXooooooooXXXXXooooooooooooooooXXXXXoooooooooXXXoooooooooooooXXoooooooooooooooooooo",
"ooooooooooooooooooXXXoooooooooooooooooooooooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"ooooooooooooooooooooooooooooooooooooooooooooooooooooooXXXooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooXoooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo",
"oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo"
};
