/*
 * Schism Tracker - a cross-platform Impulse Tracker clone
 * copyright (c) 2003-2005 Storlek <storlek@rigelseven.com>
 * copyright (c) 2005-2008 Mrs. Brisby <mrs.brisby@nimh.org>
 * copyright (c) 2009 Storlek & Mrs. Brisby
 * copyright (c) 2010-2012 Storlek
 * URL: http://schismtracker.org/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "headers.h"

#include <sys/mman.h>
#include <fcntl.h>

#include "slurp.h"

static void munmap_slurp_(slurp_t *fp)
{
	(void)munmap((void*)fp->internal.memory.data, fp->internal.memory.length);
	(void)close(fp->internal.memory.interfaces.mmap.fd);
}

int slurp_mmap(slurp_t *fp, const char *filename, uint64_t st)
{
	/* don't overflow if sizeof(uint64_t) > sizeof(size_t) */
	if (st > (uint64_t)SIZE_MAX)
		return SLURP_OPEN_IGNORE;

	int fd = open(filename, O_RDONLY);
	if (fd == -1) return SLURP_OPEN_FAIL;

	void *addr = mmap(NULL, st, PROT_READ, MAP_SHARED
#if defined(MAP_POPULATE) && defined(MAP_NONBLOCK)
		| MAP_POPULATE | MAP_NONBLOCK
#endif
#if defined(MAP_NORESERVE)
		| MAP_NORESERVE
#endif
		, fd, 0);

	if (addr == MAP_FAILED) {
		(void)close(fd);
		return (errno == ENOMEM) ? SLURP_OPEN_FAIL : SLURP_OPEN_IGNORE;
	}

	slurp_memstream(fp, addr, st);

	fp->closure = munmap_slurp_;
	fp->internal.memory.interfaces.mmap.fd = fd;

	return SLURP_OPEN_SUCCESS;
}
